﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.ComponentModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace XamDataTreeEx
{
    public class Item : INotifyPropertyChanged
    {
        public Item(string name, bool isEnabled, bool initialState)
        {
            Name = name;
            this.Children = new ObservableCollection<Item>();
            Children.CollectionChanged += Children_CollectionChanged;
            this.PropertyChanged += Difference_PropertyChanged;
            this.IsChecked = initialState;
            this.IsEnabled = isEnabled;
            this.IsExpanded = true;
        }

        /// <summary>
        /// Ctor
        /// </summary>
        /// <param name="record">CompareBilling type record</param>
        public Item(CompareBilling record)
        {
            if (!String.IsNullOrEmpty(record.SubData))
            {
                this.Name = String.Format("{0} [{1}]", record.DisplayData, record.SubData);
            }
            else
            {
                this.Name = record.DisplayData;
            }

            this.IsEnabled = record.AcceptableDifference;
            //this.IsEnabled = (record.CanAcceptChildren || record.AcceptableDifference);
            this.AcceptableDifference = record.AcceptableDifference;
            this.CanAcceptChildren = record.CanAcceptChildren;
            this.IsChange = record.IsChange;
            this.IsChecked = record.Accepted;
            this.DifferencesID = record.DifferencesID;
            this.KeyData = record.KeyData;
            this.IsExpanded = false;

            this.Children = new ObservableCollection<Item>();
            Children.CollectionChanged += Children_CollectionChanged;
            this.PropertyChanged += Difference_PropertyChanged;
        }

        #region Init

        /// <summary>
        /// Initialize object (the top most node)
        /// </summary>
        public void Initialize()
        {
            foreach (Item child in this.Children)
            {
                child.Parent = this;
                child.Initialize();
            }

            if (this.Children.Count > 0)
            {
                VerifyCheckState();
            }
        }

        #endregion

        #region Properties

        /// <summary>
        /// Is the Item Expanded or not
        /// </summary>
        public bool IsExpanded { get; set; }

        /// <summary>
        /// Is a user acceptable difference
        /// </summary>
        public bool AcceptableDifference { get; private set; }

        /// <summary>
        /// Can acceot children by clicking checkbox
        /// </summary>
        public bool CanAcceptChildren { get; private set; }

        /// <summary>
        /// Denotes Is Change
        /// </summary>
        public bool IsChange { get; private set; }

        /// <summary>
        /// Differences ID
        /// </summary>
        public int DifferencesID { get; set; }

        public bool IsDirty { get; set; }
        /// <summary>
        /// Key Data
        /// </summary>
        public string KeyData { get; private set; }

        /// <summary>
        /// The item's name
        /// </summary>
        public string Name { get; private set; }

        /// <summary>
        /// The parent item of the item (gets set when Initialized)
        /// </summary>
        Item Parent;

        /// <summary>
        /// Items collection of children items
        /// </summary>
        public ObservableCollection<Item> Children { get; private set; }

        ///// <summary>
        ///// Bool for if item is Initially Selected
        ///// </summary>
        //public bool IsInitiallySelected { get; set; }

        public bool IsEnabled { get; set; }

        /// <summary>
        /// Number of accepted differences in the items collection (or itself if no children)
        /// </summary>
        public int Accepted
        {
            get { return CountTotalAccepted(this); }
        }

        /// <summary>
        /// Total number of differences in the items collection (or itself if no children)
        /// </summary>
        public int Differences
        {
            get { return CountTotalDifferences(this); }
        }

        #region IsChecked

        private bool? _isChecked = false;
        /// <summary>
        /// Gets/sets the state of the associated UI toggle (ex. CheckBox).
        /// The return value is calculated based on the check state of all
        /// child Items.  Setting this property to true or false
        /// will set all children to the same check state, and setting it 
        /// to any value will cause the parent to verify its check state.
        /// </summary>
        public bool? IsChecked
        {
            get { return _isChecked; }
            set
            {
                this.SetIsChecked(value);//, true, true);

                if (this.AcceptableDifference)
                {
                    IsDirty = true;
                }
            }
        }

        /// <summary>
        /// Setter takes care of updating parents and children IsChecked and Count of Difference and Accepted Differences
        /// </summary>
        /// <param name="value">True if Accepted, False if not, null if Indeterminate</param>
        /// <param name="updateChildren">True to update children nodes</param>
        /// <param name="updateParent">True to update parent nodes</param>
        void SetIsChecked(bool? value)//, bool updateChildren, bool updateParent)
        {
            if (_isChecked != value)
            {
                _isChecked = value;
                FirePropertyChanged("IsChecked");
            }
        }

        /// <summary>
        /// Verifies the check state
        /// </summary>
        private void VerifyCheckState()
        {
            bool? state = null;
            for (int i = 0; i < this.Children.Count; ++i)
            {
                bool? current = this.Children[i].IsChecked;
                if (i == 0)
                {
                    state = current;
                }
                else if (state != current)
                {
                    state = null;
                    break;
                }
            }
            this.SetIsChecked(state);//, false, true);
        }

        #endregion

        #endregion

        #region Methods

        /// <summary>
        /// Count the total number of difference items within
        /// </summary>
        /// <param name="collection">Node to count differences (children done recursively)</param>
        /// <returns>Total number of differences within the node</returns>
        public int CountTotalDifferences(Item node)
        {
            int count = 0;

            if (node.Children.Count != 0)
            {
                for (int i = 0; i < node.Children.Count; i++)
                {
                    if (node.Children[i].Children.Count == 0)
                    {
                        if (node.Children[i].IsChange)
                        {
                            count++;
                        }
                    }
                    else
                    {
                        //count += node.Children[i].Differences;
                        count += CountTotalDifferences(node.Children[i]);
                    }
                }
            }
            else
            {
                if (node.IsChange)
                {
                    count = 1;
                }
            }

            return count;
        }

        /// <summary>
        /// Count the number of differences that are set to IsChecked within
        /// </summary>
        /// <param name="node">Node to count (will do children recursively)</param>
        /// <returns>Number of actual differences that are marked as accepted</returns>
        public int CountTotalAccepted(Item node)
        {
            int count = 0;

            if (node.Children.Count != 0)
            {
                for (int i = 0; i < node.Children.Count; i++)
                {
                    if (node.Children[i].Children.Count == 0)
                    {
                        if (node.Children[i].IsChecked == true)
                        {
                            count++;
                        }
                    }
                    else
                    {
                        count += node.Children[i].Accepted;
                    }
                }
            }
            else
            {
                if (IsChecked == true)
                {
                    count = 1;
                }
            }

            return count;
        }

        /// <summary>
        /// Collection changed notification
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Children_CollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
        {
            if (e.NewItems != null)
            {
                foreach (Item item in e.NewItems)
                {
                    item.PropertyChanged += Item_PropertyChanged;
                }
            }
            if (e.OldItems != null)
            {
                foreach (Item item in e.OldItems)
                {
                    item.PropertyChanged -= Item_PropertyChanged;
                }
            }
        }

        /// <summary>
        /// Updates the Accepted property anytime collection has changed
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        public void Item_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == "Accepted")
            {
                FirePropertyChanged("Accepted");
            }
        }

        /// <summary>
        /// Updates the Differences Accepted property anytime a difference is unchecked or checked
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        public void Difference_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if ((sender as Item).Children.Count == 0)
            {
                if (e.PropertyName == "IsChecked")
                {
                    FirePropertyChanged("Accepted");
                }
            }
        }

        #endregion

        #region INotifyPropertyChange

        public event PropertyChangedEventHandler PropertyChanged;

        protected void FirePropertyChanged(string propertyName)
        {
            if (PropertyChanged != null)
                PropertyChanged(this, new PropertyChangedEventArgs(propertyName));

        }

        #endregion
    }
}
